# July 2021
import numpy as np
import pandas as pd
import scipy.stats
import math

"""
Conducts GESD test for detecting outliers

See details https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm

Compliant ISO 16269-4:2010

Parameters
____________

data: array_like.

alpha: significance level, optional
    Default is 0.05.

max_outliers: int

Returns
____________

data_without_out: DataFrame

Example
____________

See example in func_examples.ipynb github

Notes
____________

r: control statistics
l: critical value

1) The data should be approximately normally distributed. 
2) It is necessary to establish the expected number of outliers
on the normal distribution schedule. 
3) max_outliers must be greater than 1
"""

def GESD(data, alpha, max_outliers):
    """
    Compute control statistics.
    """

    def test_stat(data):
        mean = np.mean(data)
        std = np.std(data)
        list_difference = []
        for i in data:
            difference = round(abs(i - mean), 4)
            list_difference.append(difference)
        max_difference = max(list_difference)
        max_ind = np.argmax(list_difference)
        r = round((max_difference / np.std(data, dtype=np.float32, ddof=1)), 4)
        return r, max_ind

    """
    Compute critical value
    """
    def critical_value(data, alpha):
        p = (1 - alpha/2)**(1 / (len(data) - 0))
        t = scipy.stats.t.ppf(p, len(data))
        l =  ((len(data) - 1) * t) / np.sqrt((len(data) - 2 + np.square(t))*(len(data) - 1))
        return l

    r = []
    l = []
    index = []
    outliers = []
    for iterations in range(0, max_outliers):
        stat, max_ind = test_stat(data)
        critical = critical_value(data, alpha)
        if stat >= critical:
            out = [e for i, e in enumerate(data) if i == max_ind]
            data = np.delete(data, max_ind)
        elif stat <= critical:
            out = 'Not outlier'
        l.append(critical)
        r.append(stat)
        outliers.append(out)
    data_without_out = pd.DataFrame({'number_of_outliers': range(0, max_outliers), 'r-stat': r, 'l-critical': l, 'index_of_outliers': list(np.hstack(outliers))})
    print(data_without_out)


# September 2021
import numpy as np
import pandas as pd
import scipy.stats
import math
import pkgutil

"""
Modified boxplot where the lower and upper quarters are used.

See details https://www.itl.nist.gov/div898/handbook/eda/section3/boxplot.htm

Compliant ISO 16269-4:2010

Parameters
____________

data: array_like.

alpha: significance level (0.05 or 0.01), optional
    Default is 0.05.

Returns
____________

outliers: array 
        Outliers in data.

Example
____________

See example in func_examples.ipynb github

Notes
____________

1) For normal distribution only
2) No preliminary estimate of the amount of emissions required
3) The table shows values only for samples from 9 to 500 values
"""




def boxplot(data, alpha = 0.05):
    data_boxplot = np.array(data)
    n = len(data_boxplot)
    criterion = n / 4
    """
    """
    i = np.modf(criterion)[1]
    if np.modf(criterion)[0] == 0:
        x_l = (data_boxplot[int(i-1)] + data_boxplot[int(i)]) / 2
        x_u = (data_boxplot[int(n-i-1)] + data_boxplot[int(n-i)]) / 2
    else:
        x_l = data_boxplot[int(i+1)]
        x_u = data_boxplot[int(n-i-1)]
    """
    """
    data_pac = pkgutil.get_data(__name__, "datasets/coef.xlsx")
    coef = pd.read_excel(data_pac)
    coef['a'] = coef['a'].astype('str')
    coef['mod'] = coef['mod'].astype('str')
    coef['mod'] = coef['mod'].astype('str')
    coef = coef.set_index(['a', 'mod'])
    alpha = alpha
    if alpha == 0.05:
        y = np.mod(n, 4)
        if y == 0:
            b0, b1, b2, b3, b4 = coef.xs('0.05').xs('0')[0:5]

        if y == 1:
            b0, b1, b2, b3, b4 = coef.xs('0.05').xs('1')[0:5]

        if y == 2:
            b0, b1, b2, b3, b4 = coef.xs('0.05').xs('2')[0:5]

        if y == 3:
            b0, b1, b2, b3, b4 = coef.xs('0.05').xs('3')[0:5]
            
    else:
        y = np.mod(n, 4)
        if y == 0:
            b0, b1, b2, b3, b4 = coef.xs('0.01').xs('0')[0:5]
            
        if y == 1:
            b0, b1, b2, b3, b4 = coef.xs('0.01').xs('1')[0:5]
            
        if y == 2:
            b0, b1, b2, b3, b4 = coef.xs('0.01').xs('2')[0:5]
            
        if y == 3:
            b0, b1, b2, b3, b4 = coef.xs('0.01').xs('3')[0:5]
            
    k = np.exp(b0 + b1*np.log(n) + b2*((np.log(n))**2) + b3*((np.log(n))**3) + b4*((np.log(n))**4)) 
    L_f = x_l - (k * (x_u - x_l))
    U_f = x_u + (k * (x_u - x_l)) 
    outliers = data_boxplot[(data_boxplot < L_f) | (data_boxplot > U_f)]
    return outliers  


# September 2021
import numpy as np
import pandas as pd
import scipy.stats
import math
import pkgutil

"""
Greenwood's criterion for detecting outliers in data.

Compliant ISO 16269-4:2010

Parameters
____________

data: array_like.

alpha: significance level (0.01 or 0.025), optional
    Default is 0.025.

Returns
____________

Greenwood's criterion, Critical value maximum, Critical value minimum, Result.

Example
____________

See example in func_examples.ipynb github

Notes
____________

The sample must obey an exponential distribution. The table shows the values for the significance level 0.025 and 0.01.
"""

def greenwood(data, alpha=0.025):
    data_pac = pkgutil.get_data(__name__, "datasets/greenwood_1.xlsx")
    table_criterion = pd.read_excel(data_pac)
    low = "low " + str(alpha)
    up = "up " + str(alpha)
    n = len(data)
    n1 = n -1
    critical_value = table_criterion[table_criterion['n'] == n1][[low, up]]
    a = np.min(data)
    p = []
    n = len(data)
    for i in data:
        o = (i - a)**2
        p.append(o)
    summa = np.sum(p)
    z = []
    for ix in data:
        z1 = ix - a
        z.append(z1)
    summa1 = np.sum(z)
    b = (summa1)**2
    greenwood_criterion = summa / b

    if greenwood_criterion > critical_value.values[0,1]:
        outliers = 'Outliers among maximum values'
    if greenwood_criterion < critical_value.values[0,0]:
        outliers = 'Outliers among minimum or/and maximum values'
    
    print("Greenwood's criterion:{f}".format(f = greenwood_criterion))
    print()
    print("Critical value maximum: {f}".format(f = critical_value.values[0,1]))
    print()
    print("Critical value minimum: {f}".format(f = critical_value.values[0,0]))
    print()
    print('Result: {f}'.format(f = outliers))


# July 2021
import numpy as np
import pandas as pd
import scipy.stats
import math
import pkgutil

"""
Compute trimmed mean.

See details https://en.wikipedia.org/wiki/Truncated_mean

Compliant ISO 16269-4:2010

Parameters
____________

data: array_like.

Returns
____________

data_mean: DataFrame.
        Contains percentage outliers and trimmed mean.

Example
____________

See example in func_examples.ipynb github

Notes
____________

For symmetric distributions only.
"""

def trim_mean(data):
    alpha = [0.05, 0.10, 0.15, 0.18, 0.20]
    len_data = len(data)
    mean_column = []
    for i in alpha:
        a = 1 / (len_data * (1-(2*i)))
        r = math.floor(i * len_data)
        g = i * len_data - r
        x_r = data[r]
        x_n = data[len_data-r-1]
        x_i = data[(r+2-1):(len_data-r-1)]
        b = (1 - g) * (x_r + x_n) + np.sum(x_i)
        xmean = (1 / (len_data * (1-(2*i)))) * ((1 - g) * (x_r + x_n) + np.sum(x_i))
        mean_column.append(xmean)
    data_mean = pd.DataFrame({"per_outliers": [0.05, 0.10, 0.15, 0.18, 0.20], "trim_mean": mean_column})
    print(data_mean)
    print()
    print('Mean for whole data: {m}'.format(m=np.mean(data)))
    print()
    print('Median for whole data: {md}'.format(md=np.median(data)))


# September 2021
import numpy as np
import pandas as pd
import scipy.stats
import math
import pkgutil

"""
Cochran's test for detecting outliers in variances.

See details https://www.itl.nist.gov/div898/software/dataplot/refman1/auxillar/cochvari.htm

Complaint ISO 16269-4:2010

Parameters
____________

data: array_like.
    Obtained variances.

n: int.
    Amount observations from wich variances was computed.

alpha: significance level (0.05, 0.01 or 0.001), optional
    Default is 0.05.

Returns
____________

Critical_value, Cochran's criterion, Outliers.

Example
____________

See example in func_examples.ipynb github

Notes
____________

The table shows the values for the significance level 0.05, 
0.01, 0.001 for the number of variances from 2 to 40. 
In this case, the variances themselves should be determined 
from observations in the amount from 2 to 10.
"""

def cochran(data, n, alpha=0.05):
    data_pac = pkgutil.get_data(__name__, "datasets/cochran.xlsx") 
    cochran = pd.read_excel(data_pac)
    cochran = cochran.set_index(['alpha', 'p'])
    
    amount_variances = len(data) 
    critical_value = cochran.xs(alpha).xs(amount_variances)[n]
    c = np.max(data) / np.sum(data)
    if c > critical_value:
        outliers = np.max(data)
    else:
        outliers = 'No'
    print('Critical_value:{f}'.format(f=critical_value))
    print()
    print("Cochran's criterion: {f}".format(f=c))
    print()
    print("Outliers:{f}".format(f=outliers))